import react from 'react';
import { useAtom } from 'jotai';

import { TileKeyObject } from './types';

import { coordinateSystemAtom } from './Map';
import TileSet from './TileSet';
import _, { floor, range } from 'lodash';
import tileUri from './uris';

export interface LayerStackProperties {
  /**
   * A key identifying the initial top left tile
   */
  keyObject: TileKeyObject;
  /**
   * Number of {@link components/map/TiledLayer!TiledLayer}.
   */
  numberOfTiledLayers?: number;
}

/**
 *
 * @param props
 * @returns A stack of layers embedded in an SVG element
 *
 * This component does the conversion between the {@link components/map/Map!CoordinateSystem} stored
 * in the {@link components/map/Map!coordinateSystemAtom} atom and the {@link components/map/TiledLayer!TiledLayer}
 * components which units are in tiles.
 *
 */
export const LayerStack: react.FC<LayerStackProperties> = (
  props: LayerStackProperties
) => {
  const [coordinateSystem] = useAtom(coordinateSystemAtom);

  const numberOfTiledLayers =
    props.numberOfTiledLayers === undefined ? 1 : props.numberOfTiledLayers;

  const activeTiledLayer = Math.min(
    Math.max(
      Math.round(Math.log2(coordinateSystem.zoom)) +
        Math.floor(numberOfTiledLayers / 2),
      0
    ),
    numberOfTiledLayers - 1
  );

  const viewPort = {
    topLeft: {
      x:
        props.keyObject.x +
        Math.floor(-coordinateSystem.shift.x / coordinateSystem.zoom / 256),
      y:
        props.keyObject.y +
        Math.floor(-coordinateSystem.shift.y / coordinateSystem.zoom / 256),
    },
    bottomRight: {
      x:
        props.keyObject.x +
        Math.ceil(
          (-coordinateSystem.shift.x + window.innerWidth) /
            coordinateSystem.zoom /
            256
        ) -
        1,
      y:
        props.keyObject.y +
        Math.ceil(
          (-coordinateSystem.shift.y + window.innerHeight) /
            coordinateSystem.zoom /
            256
        ) -
        1,
    },
  };

  const getTileSet = (i: number) => {
    const relativeZoomLevel = i - Math.floor(numberOfTiledLayers / 2);
    const zoom = 2 ** relativeZoomLevel;
    const origin = {
      x: props.keyObject.x * zoom,
      y: props.keyObject.y * zoom,
    };
    const keyObject = {
      provider: props.keyObject.provider,
      zoomLevel: props.keyObject.zoomLevel + relativeZoomLevel,
      x: Math.floor(origin.x),
      y: Math.floor(origin.y),
    };
    const shift = {
      x: origin.x - floor(origin.x),
      y: origin.y - floor(origin.y),
    };
    const key = tileUri({
      provider: keyObject.provider,
      zoomLevel: keyObject.zoomLevel,
    });

    return (
      <g
        transform={`scale(${256 / zoom}) translate(${shift.x - keyObject.x}, ${
          shift.y - keyObject.y
        })`}
        key={key}
        data-testid={key}
      >
        <TileSet
          key={key}
          keyObject={keyObject}
          viewPort={i === activeTiledLayer ? viewPort : undefined}
        />
      </g>
    );
  };

  //  console.log(`tiledLayers: ${JSON.stringify(tiledLayers)}`);

  return (
    <svg
      width={window.innerWidth}
      height={window.innerHeight}
      data-testid='layer-stack'
    >
      <g
        transform={`translate(${coordinateSystem.shift.x}, ${coordinateSystem.shift.y}) scale(${coordinateSystem.zoom})`}
      >
        {
          // Tiled layers with less detail
          range(0, activeTiledLayer).map(getTileSet)
        }
        {
          // Tiled layers with more details
          range(numberOfTiledLayers - 1, activeTiledLayer, -1).map(getTileSet)
        }
        {
          // And the active one
          getTileSet(activeTiledLayer)
        }
      </g>
    </svg>
  );
};

export default LayerStack;
